//This work is licensed under a Creative Commons Attribution 3.0 Unported License.
//Original version by Daniel Davis, 2013
//danieldavis.com

public abstract class ch_abstract
{
  protected String[] words = new String[]{"ABCDE", "FGHIJK", "LMNOP", "QRSTU", "VWXYZ"};
  protected int font_size = 160;
  protected float step_size = 150;
  RFont font = new RFont(font_family, font_size, RFont.CENTER);
  
  
  public abstract void alphabet(String fileName);
  
  protected void setup()
  {
    size(800,800);
    background( white);
  }
  
  //////////////////////////////////////////////////
  // HELPERS
  //////////////////////////////////////////////////
  
  
  /**
   * Returns a mix of the blue and red colors.
   * When proportion =
   * 0: totally red
   * 1: totally blue
   * 0.5: purple
   * @param proportion
   * @return
   */
  public color colorMix(float proportion)
  {
    return lerpColor(red, blue, proportion);
  }
  
   /**
   * Allows a curve vertex to be made from RPoint geometry.
   * @param point
   */
  public void curveVertexFromPoint(RPoint point)
  {
    curveVertex(point.x, point.y);
  }
  
  public void lineFromPoint(RPoint start, RPoint end)
  {
    line(start.x, start.y, end.x, end.y);
  }
  
  public void beginPDFRecord(String recordTo)
  {
    if(recordTo != "")
    {
      beginRecord(PDF, "pdf/" + recordTo + ".pdf");
      drawSpacer();
    }
  }
  
  public void endPDFRecord()
  {
    endRecord();
  }
      
  
  /**
   * Draws two tiny little lines in the top and bottom corners to ensure the PDF is the correct size.
   */
  private void drawSpacer()
  {
    stroke(255);
    noFill();
    strokeWeight(0.1f);
  
    line(0, 0, 1, 0);
    line(width, height, width - 1, height);
  }
  
  
  
  //////////////////////////////////////////////////
  // GEOMETRY
  //////////////////////////////////////////////////
  
  /**
   * Returns a point on a circle 
   * 
   * @param center
   * @param radius 
   * @param angle = in radians
   */
  protected RPoint pointOnCircle(RPoint center, float radius, float angle)
  {
    float x = center.x + (radius * (float)Math.cos(angle));
    float y = center.y + (radius * (float)Math.sin(angle));
    return new RPoint(x,y);
  }
  
  
  /**
   * Works out if a line intersects the boundry of a shape.
   * 
   * @param shape = a collection of points that define the shape, where each points goes sequentually around the boundry of the shape
   */
  protected boolean lineIntersectsShape(RPoint p1, RPoint p2, RPoint[] shape)
  {
    for(int i = 0; i < shape.length - 1; i++ )
    {
      RPoint p3 = shape[i];
      RPoint p4 = shape[i+1];
      
      //check they are not the same points.
      if(p1 == p3) continue;
      if(p2 == p3) continue;
      if(p1 == p4) continue;
      if(p2 == p4) continue;
      
      if(lineIntersection(p1, p2, p3, p4))
      {
        return true;
      }
    }
    return false;
  }
  
  
  
  /**
   * To determine if a line is inside the shape. We see if the center of the line is inside the shape. 
   * Draw a new line from the center to some point in the distance.
   * Then cound how many times this new line intersects with the boundry of the shape.
   * If it is an even number, the line is outside the shape, returns false.
   * if it is an odd number, the line is inside the shape, returns true.
   * @param x1
   * @param y1
   * @param x2
   * @param y2
   * @param shape
   * @return
   */
  protected boolean isLineInShape(RPoint p1, RPoint p2, RPoint[] shape)
  {
    float midX = p1.x + (p2.x-p1.x) * 0.5f;
    float midY = p1.y + (p2.y-p1.y) * 0.5f;
    RPoint mid = new RPoint (midX, midY);
    
    int count = 0;
    
    for(int i = 0; i < shape.length - 1; i++ )
    {
      if(lineIntersection(mid , new RPoint(10000,10000) ,shape[i], shape[i+1]))
      {
        count++;
      }
    }
    
    if(count % 2 == 0)
    {
      return false;
    }
    return true;
  }
  
    /**
   * x1,y1,x2,y2 define one line
   * x3,y3,x4,y4 define another line.
   * @param x1
   * @param y1
   * @param x2
   * @param y2
   * @param x3
   * @param y3
   * @param x4
   * @param y4
   * @return true if the lines intersect within the bounds of the line (not infinite rays).
   * Otherwise return false.
   */
  protected boolean lineIntersection(RPoint p1, RPoint p2, RPoint p3, RPoint p4)
  {
    float bx = p2.x - p1.x; 
    float by = p2.y - p1.y; 
    float dx = p4.x - p3.x; 
    float dy = p4.y - p3.y;
    float b_dot_d_perp = bx * dy - by * dx;
    if(b_dot_d_perp == 0) {
      return false;
    }
    float cx = p3.x - p1.x;
    float cy = p3.y - p1.y;
    float t = (cx * dy - cy * dx) / b_dot_d_perp;
    if(t < 0 || t > 1) {
      return false;
    }
    float u = (cx * by - cy * bx) / b_dot_d_perp;
    if(u < 0 || u > 1) { 
      return false;
    }
    return true;
  }
}
